/*
 *  forza_pcm.c is sound driver for custom fpga
 *
 *  Copyright 2013,2014 Sony Corporation
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  version 2 of the  License.
 *
 *  THIS  SOFTWARE  IS PROVIDED   ``AS  IS'' AND   ANY  EXPRESS OR IMPLIED
 *  WARRANTIES,   INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO  EVENT  SHALL   THE AUTHOR  BE    LIABLE FOR ANY   DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED   TO, PROCUREMENT OF  SUBSTITUTE GOODS  OR SERVICES; LOSS OF
 *  USE, DATA,  OR PROFITS; OR  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
 */

#include <linux/module.h>   // required by all modules
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/sched.h>
#include <linux/moduleparam.h>
#include <linux/slab.h>
#include <sound/core.h>
#include <asm/cacheflush.h>
#include <sound/initval.h>
#include <sound/info.h>
#include <asm/delay.h>
#include "forza.h"
#include "forza_lib.h"
#include "forza_pcm.h"
#include "forza_defs.h"
#include "dsp_ops.h"
#include "forza_audio_controller.h"


/*
 * PCM PLAYBACK
 */
static struct snd_pcm_hardware snd_forza_playback_hw = {
  .info =
  (SNDRV_PCM_INFO_MMAP |
   SNDRV_PCM_INFO_RESUME |
   SNDRV_PCM_INFO_PAUSE | 
   SNDRV_PCM_INFO_INTERLEAVED |
   SNDRV_PCM_INFO_BLOCK_TRANSFER |
   SNDRV_PCM_INFO_MMAP_VALID),
  .formats      = SNDRV_PCM_FMTBIT_S32_LE,
  .rates        = SNDRV_PCM_RATE_CONTINUOUS | SNDRV_PCM_RATE_8000_192000,
  //  .rates        = SNDRV_PCM_RATE_CONTINUOUS | SNDRV_PCM_RATE_8000_384000,
  .rate_min     = 8000,
  .rate_max     = 192000,
  .channels_min = 2,
  .channels_max = 2,
  .buffer_bytes_max = 512*1024,
  .period_bytes_min = 8*1024,
  .period_bytes_max = 64*1024,
  .periods_min  = 8,
  .periods_max  = 8,
  .fifo_size    = 1024*32,
};

/*
 * DSD PLAYBACK
 */
static struct snd_pcm_hardware snd_forza_dsd_playback_hw = {
  .info =
  (SNDRV_PCM_INFO_MMAP |
   SNDRV_PCM_INFO_RESUME |
   SNDRV_PCM_INFO_PAUSE | 
   //   SNDRV_PCM_INFO_NONINTERLEAVED |
   SNDRV_PCM_INFO_INTERLEAVED |
   SNDRV_PCM_INFO_BLOCK_TRANSFER |
   //   SNDRV_PCM_INFO_SYNC_START |
   SNDRV_PCM_INFO_MMAP_VALID),
  .formats      = SNDRV_PCM_FMTBIT_S32_LE,
  .rates        = SNDRV_PCM_RATE_CONTINUOUS | SNDRV_PCM_RATE_8000_192000,
  .rate_min     = 88200, 
  .rate_max     = 192000,
  .channels_min = 2,
  .channels_max = 2,
  .buffer_bytes_max = 512*1024,
  .period_bytes_min = 8*1024,
  .period_bytes_max = 64*1024,
  .periods_min  = 8,
  .periods_max  = 8,
  .fifo_size    = 32*1024,
};


static void snd_forza_pcm_free_substream(struct snd_pcm_runtime *runtime)
{
  FORZA_DEBUG_LOG("kfree:substream_free\n");
  kfree(runtime->private_data);
}

// PCM open callback
static int snd_forza_playback_open(struct snd_pcm_substream *substream)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_pcm_runtime *runtime = substream->runtime;
  struct snd_forza_pcm *fpcm;
  int err;
  FORZA_DEBUG_LOG("PLAYBACK_OPEN:PCM >>>>>");
  fpcm = kzalloc(sizeof( *fpcm), GFP_KERNEL);
  if(fpcm == NULL){
    FORZA_ERROR_LOG("Can't allocate snd_forza_pcm");
    return -ENOMEM;
  }

  chip->cra_ops->ena_pcie_irq(chip, SCDMA_IRQ);
  substream->runtime->private_data = fpcm;

  fpcm->chip      = chip;
  fpcm->substream = substream;
  fpcm->type      = PLAYBACK_MUSIC;
  chip->substream = substream;
  chip->IsPlayReq = DEVICE_ENABLE;
  runtime->private_data = fpcm;
  runtime->hw = snd_forza_playback_hw;
  runtime->private_free = snd_forza_pcm_free_substream;
  err= snd_pcm_hw_constraint_minmax(runtime, SNDRV_PCM_HW_PARAM_BUFFER_SIZE,0, 128*1024);
  if(err < 0){
    FORZA_ERROR_LOG("Can't hw_constraint snd_forza_pcm");
    return err;
  }
  snd_pcm_set_sync(substream);
  forza_print_substream_info(chip->substream);
  chip->audio_ops->mute(chip,DEVICE_ENABLE);
  return 0;
}


// PCM close callback
static int snd_forza_playback_close(struct snd_pcm_substream *substream)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  chip->cra_ops->dis_pcie_irq(chip, SCDMA_IRQ);
  chip->dma_ops->reset(chip);
  chip->IsPlayReq = DEVICE_DISABLE;
  FORZA_DEBUG_LOG("PLAYBACK_CLOSE <<<<<");
  return 0;
}

// PCM hw_params call back
static int snd_forza_playback_hw_params(struct snd_pcm_substream *substream,
					struct snd_pcm_hw_params *hw_params)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_pcm_runtime *runtime = substream->runtime;
  struct snd_forza_pcm *fpcm = runtime->private_data;
  int err;

  FORZA_DEBUG_LOG("PLAYBACK_HW_PARAMS\n");
  err = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params));
  if(err <0){
    FORZA_ERROR_LOG("PLAYBACK_HW_PARAMS:malloc_pages");
    return err;
  }
  fpcm->dma_addr        = substream->runtime->dma_addr;
  fpcm->dma_bytes       = params_buffer_bytes(hw_params);
  if(FORZA_DATA_PREOUT == DEVICE_ENABLE){
    chip->preout_size = 0;
    memset(chip->preout, 0 ,FORZA_DATA_PREOUT_SIZE);
  }
  chip->audio_ops->formatter(chip,DEVICE_DISABLE);
  return 0;
}

//PCM  hw_free callback
static int snd_forza_playback_hw_free(struct snd_pcm_substream *substream)
{
  int err = snd_pcm_lib_free_pages(substream);
  if(err < 0) return err;
  return 0;
}

//PCM prepare callback
static int snd_forza_playback_prepare(struct snd_pcm_substream *substream)
{
  unsigned int fifo = 0;
  unsigned int fifo_cnt = 0;
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_pcm_runtime *runtime = substream->runtime;
  struct snd_forza_pcm *fpcm  = runtime->private_data;
  // DSP RESET
  fpcm->period_size = runtime->period_size;
  fpcm->buffer_size = runtime->buffer_size;
  fpcm->period_pos  = 0;
  fpcm->periods     = runtime->periods;
  fpcm->dma_ptr     = 0;
  fpcm->buf_periods  = runtime->dma_bytes/runtime->buffer_size;
  chip->int_cnt = 0;
  chip->dma_ops->reset(chip);
  fpcm->dma_area    = runtime->dma_area;
  fpcm->dma_addr    = runtime->dma_addr;
  fpcm->period_size = runtime->period_size;
  fpcm->cur_ptr  = 0;
  chip->cra_ops->set_txa(chip, fpcm->dma_addr);
  // DMA SETUP
  chip->dma_ops->setup(chip,
		       0, 
		       fpcm->dma_ptr,
		       fpcm->period_size);
  // AUDIO SETTINGS
  switch(runtime->rate){
  case 44100:
  case 48000:
    fifo = FORZA_FIFO_48K_SIZE;
    break;
  case 88200:
  case 96000:
    fifo = FORZA_FIFO_96K_SIZE;
    break;
  case 176400:
  case 192000:
    fifo = FORZA_FIFO_192K_SIZE;
    break;
  }
  
  fifo_cnt = (fifo / fpcm->period_size);
  FORZA_DEBUG_LOG(">>DelayCnt(%d)\n",fifo_cnt);
  // DELAY SETUP for FIFO
  chip->dmfi.msdelay = FORZA_HW_FIFO_PCM_CHARGE_DELAY;
  chip->dmfi.times   = fifo_cnt +1;
  chip->dmfi.start_pos = chip->dmfi.times ;
  // AUDIO SET
  if(chip->IsPlayReq == DEVICE_DISABLE){
    FORZA_DEBUG_LOG("PLAYBACK_PREPARE(PASS)");
  }else{
  // CRA
    FORZA_DEBUG_LOG("PLAYBACK_PREPARE");
    chip->audio_ops->set_pcm(chip, runtime->rate, 32, FORZA_HEQ_MODE_LPCM, 0);
  }
  return 0;
}// PCM


//// PCM  trigger callback
static int snd_forza_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_forza_pcm *fpcm = substream->runtime->private_data;
  switch (cmd){
  case SNDRV_PCM_TRIGGER_START:
    FORZA_INFO_LOG("TRG.pcm.start\n");
    fpcm->running = 1;
    fpcm->pointer_called_cnt = 0;
    fpcm->irq_called_cnt = 0;
    fpcm->dma_period_cnt = 0;
    fpcm->sum =0;
    chip->dac_ops->output(chip, DEVICE_ENABLE);
    forza_print_runtime_info(substream->runtime,"START");

    /*
     * PLAYBACK_TRIGGER_DELAY
     */
    if(PLAYBACK_TRIGGER_DELAY == 0){
      FIFO_CONTROL_REG_W(chip,1);
      chip->dma_ops->start(chip);
      FORZA_DEBUG_LOG("DMA Begining.....");
    }
    break;
  case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
    fpcm->running = 1;
    chip->audio_ops->formatter(chip,DEVICE_ENABLE);
    FORZA_DEBUG_LOG("PAUSE_RELEASE.....");
    break;
  case SNDRV_PCM_TRIGGER_RESUME:
    fpcm->running = 1;
    FORZA_DEBUG_LOG("RESUME....");
    break;
  case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
    fpcm->running = 0;
    FORZA_DEBUG_LOG("PAUSE_PUSH.....");
    break;
  case SNDRV_PCM_TRIGGER_SUSPEND:
    fpcm->running = 0;
    FORZA_DEBUG_LOG("SUSPEND.....");
    break;
  case SNDRV_PCM_TRIGGER_STOP:
    chip->audio_ops->mute(chip,DEVICE_ENABLE);
    chip->IsPlayReq = DEVICE_DISABLE;
    FORZA_INFO_LOG("TRG.pcm.stop\n");
    udelay(800);
    fpcm->running = 0;
    fpcm->irq_called_cnt = 0;
    fpcm->dma_period_cnt = 0;
    if(FORZA_DATA_PREOUT == DEVICE_ENABLE){
      forza_preout_data_to_file(chip);
    }
    break;
  default:
    FORZA_DEBUG_LOG(">>DEFAULT\n");
    return -EINVAL;
  }
  return 0;
}

// PCM
void snd_forza_playback_interrupt(struct snd_forza *forza){
  
  struct snd_pcm_substream *substream = forza->substream;
  struct snd_pcm_runtime *runtime = forza->substream->runtime;
  struct snd_forza_pcm *fpcm = runtime->private_data;
  if(substream == NULL){
    FORZA_DEBUG_LOG("STREAM is NULL\n");
    return ;
  }
  fpcm->irq_called_cnt +=1;
  fpcm->dma_period_cnt +=1;

  if(fpcm->period_pos == (substream->runtime->periods -1)){
    fpcm->period_pos = 0;
  }else{
    fpcm->period_pos +=1;
  }
  fpcm->sum += fpcm->period_size;
  spin_lock(&forza->flock);

  if(fpcm->irq_called_cnt == (forza->dmfi.start_pos )){
    forza->audio_ops->formatter(forza,DEVICE_ENABLE);
  }
  if(fpcm->irq_called_cnt  == (forza->dmfi.start_pos   +3)){
    forza->audio_ops->mute(forza,DEVICE_DISABLE);
   } 

  fpcm->dma_ptr += substream->runtime->period_size;
  if(fpcm->dma_ptr == fpcm->dma_bytes){
    fpcm->dma_ptr =0;
  }else if(fpcm->dma_ptr > fpcm->dma_bytes){
    FORZA_INFO_LOG("DMA POINTER to OVER!! \n");
  }
  forza->int_cnt += 1;

  if(DMA_DESTINATION == FORZA_ONCHIP_MEM_BUF)
     forza_print_onchip_mem_info(forza, "DMA",0x0/* fpcm->dma_ptr*/,fpcm->period_size,2);

#if IRQ_LOOP_MAX != 0
    if(forza->int_cnt >= IRQ_LOOP_MAX){
       PCIE_IRQ_ENABLE_W(forza,0);
       FORZA_INFO_LOG("Loop Limit is reached!! (%d)\n",forza->int_cnt);
       spin_unlock(&forza->flock);
       return;
    }
#endif

  if(fpcm->running == 1){
    forza->dma_ops->setup(forza, 0 , fpcm->dma_ptr ,  fpcm->period_size);
    forza->dma_ops->start(forza);
  }

  if(FORZA_DATA_PREOUT == DEVICE_ENABLE){
    if(forza->preout_size < FORZA_DATA_PREOUT_SIZE){
      forza->preout_size += 256;
      FORZA_INFO_LOG(">>preout \n");
      memcpy((forza->preout+forza->preout_size), fpcm->dma_area+fpcm->dma_ptr, 256);
    }
  }

  if((fpcm->irq_called_cnt % (fpcm->periods * fpcm->buf_periods)) == 0){
    if((fpcm->running == 1) && (fpcm->irq_called_cnt < 100)){
      forza_print_mem_dump(substream->runtime->dma_area + fpcm->dma_ptr ,0x10,"NEXT" );
    }
  }
  spin_unlock(&forza->flock);
#if 0
  if(fpcm->irq_called_cnt >= 0){
    snd_pcm_period_elapsed(substream);
  }else{
      FORZA_INFO_LOG("IRQ_-1 NOT ELAPSED!!\n");
  }
#else // the above comparison is meaningless because, unsigned irq_called_cnt
      // won't be below zero. Coverity warned.
  snd_pcm_period_elapsed(substream);
#endif
  return;
}

// PCM & DSD pointer callback
static snd_pcm_uframes_t snd_forza_playback_pointer(struct snd_pcm_substream *substream)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_forza_pcm *fpcm = substream->runtime->private_data;
  unsigned int ptr;
  snd_pcm_uframes_t frames;

  spin_lock(&chip->flock);
  if(! fpcm->running)  return 0;
  fpcm->pointer_called_cnt += 1;

  if(PLAYBACK_TRIGGER_DELAY > 0){
    if(fpcm->pointer_called_cnt == (PLAYBACK_TRIGGER_DELAY-1)){
      FORZA_DEBUG_LOG("PLAYBACK_POINTER:(%d)\n",fpcm->pointer_called_cnt);
      spin_unlock(&chip->flock);
      return 0;
    }
    if(fpcm->pointer_called_cnt == PLAYBACK_TRIGGER_DELAY){
      FIFO_CONTROL_REG_W(chip,1); // FIFO Clear
      forza_print_mem_dump(substream->runtime->dma_area + fpcm->dma_ptr ,0x30,"NEXT" );
      if(FORZA_DATA_PREOUT == DEVICE_ENABLE){
	chip->preout_size = 256;
	memcpy(chip->preout, fpcm->dma_area, 256);
      }
      chip->dma_ops->start(chip);
      fpcm->period_pos +=1;
      FORZA_DEBUG_LOG("DMA Begining.....");
      spin_unlock(&chip->flock);
      return 0;
    }
  }else{
    if(fpcm->pointer_called_cnt == 1){
      forza_print_mem_dump(substream->runtime->dma_area + fpcm->dma_ptr ,0x30,"NEXT" );
    }
  }

  if(fpcm->irq_called_cnt == 0){
    spin_unlock(&chip->flock);
    return 0; 
  }
  ptr = fpcm->dma_ptr;
  spin_unlock(&chip->flock);
  frames = bytes_to_frames(substream->runtime, ptr);
  return frames;
}


// DSD open callback
static int snd_forza_dsd_playback_open(struct snd_pcm_substream *substream)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_pcm_runtime *runtime = substream->runtime;
  struct snd_forza_pcm *fpcm;
  int err;
  FORZA_DEBUG_LOG("PLAYBACK_OPEN:DSD >>>>>");
  fpcm = kzalloc(sizeof( *fpcm), GFP_KERNEL);
  if(fpcm == NULL){
    FORZA_ERROR_LOG("Can't allocate snd_forza_pcm");
    return -ENOMEM;
  }

  chip->cra_ops->ena_pcie_irq(chip, SCDMA_IRQ);
  substream->runtime->private_data = fpcm;

  fpcm->chip      = chip;
  fpcm->substream = substream;
  fpcm->type      = PLAYBACK_MUSIC;
  chip->substream = substream;
  chip->IsPlayReq = DEVICE_ENABLE;
  runtime->private_data = fpcm;
  runtime->hw = snd_forza_dsd_playback_hw;
  runtime->private_free = snd_forza_pcm_free_substream;

  err= snd_pcm_hw_constraint_minmax(runtime, SNDRV_PCM_HW_PARAM_BUFFER_SIZE,0, 128*1024);
  if(err < 0){
    FORZA_ERROR_LOG("Can't hw_constraint snd_forza_pcm");
    return err;
  }
  snd_pcm_set_sync(substream);
  forza_print_substream_info(chip->substream);
  chip->audio_ops->mute(chip,DEVICE_ENABLE);
  return 0;
}


//DSD  close callback
static int snd_forza_dsd_playback_close(struct snd_pcm_substream *substream)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  chip->cra_ops->dis_pcie_irq(chip, SCDMA_IRQ);
  chip->dma_ops->reset(chip);
  chip->IsPlayReq = DEVICE_DISABLE;
  FORZA_DEBUG_LOG("PLAYBACK_CLOSE <<<<<");
  return 0;
}

//DSD  hw_params call back
static int snd_forza_dsd_playback_hw_params(struct snd_pcm_substream *substream,
					struct snd_pcm_hw_params *hw_params)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_pcm_runtime *runtime = substream->runtime;
  struct snd_forza_pcm *fpcm = runtime->private_data;
  int err;

  FORZA_DEBUG_LOG("PLAYBACK_HW_PARAMS\n");
  err = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params));
  if(err <0){
    FORZA_ERROR_LOG("PLAYBACK_HW_PARAMS:malloc_pages");
    return err;
  }

  fpcm->dma_addr        = substream->runtime->dma_addr;
  fpcm->dma_bytes       = params_buffer_bytes(hw_params);
  chip->audio_ops->formatter(chip,DEVICE_DISABLE);
  return 0;
}

//DSD hw_free callback
static int snd_forza_dsd_playback_hw_free(struct snd_pcm_substream *substream)
{
  int err;
  FORZA_DEBUG_LOG("substream_free\n");
  err = snd_pcm_lib_free_pages(substream);
  if(err < 0){
    return err;
  }
  return 0;
}

//DSD prepare callback
static int snd_forza_dsd_playback_prepare(struct snd_pcm_substream *substream)
{
  unsigned int fifo;
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_pcm_runtime *runtime = substream->runtime;
  struct snd_forza_pcm *fpcm  = runtime->private_data;
  fpcm->period_size = runtime->period_size;
  fpcm->buffer_size = runtime->buffer_size;
  fpcm->period_pos  = 0;
  fpcm->periods     = runtime->periods;
  fpcm->dma_ptr     = 0;
  fpcm->buf_periods  = runtime->dma_bytes/runtime->buffer_size;
  chip->int_cnt = 0;
  chip->dma_ops->reset(chip);

  fpcm->dma_area    = runtime->dma_area;
  fpcm->dma_addr    = runtime->dma_addr;
  fpcm->period_size = runtime->period_size;
  fpcm->cur_ptr  = 0;

  chip->cra_ops->set_txa(chip, fpcm->dma_addr);
  
  chip->dma_ops->setup(chip,
		       0, 
		       fpcm->dma_ptr,
		       fpcm->period_size);

  switch(runtime->rate){
  case 88200:
  case 96000:
    fifo = FORZA_FIFO_96K_SIZE;
    break;
  case 176400:
  case 192000:
    fifo = FORZA_FIFO_192K_SIZE;
  break;
  }

  chip->dmfi.msdelay = FORZA_HW_FIFO_DSD_CHARGE_DELAY;
  chip->dmfi.times   = 3;
  chip->dmfi.start_pos = chip->dmfi.times ;

  if(chip->IsPlayReq == DEVICE_DISABLE){
    FORZA_DEBUG_LOG("PLAYBACK_PREPARE(PASS)");
    return 0;
  }else{
    FORZA_DEBUG_LOG("PLAYBACK_PREPARE");
    chip->audio_ops->set_dsd(chip, runtime->rate );
  }
  return 0;
}

//DSD trigger callback
static int snd_forza_dsd_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  struct snd_forza_pcm *fpcm = substream->runtime->private_data;
  FORZA_DEBUG_LOG("DSD_TRIGGER");
  switch (cmd){
  case SNDRV_PCM_TRIGGER_START:
    FORZA_INFO_LOG("TRG.dsd.start\n");
    fpcm->running = 1;
    fpcm->pointer_called_cnt = 0;
    fpcm->irq_called_cnt = 0;
    fpcm->dma_period_cnt = 0;
    fpcm->sum =0;

    chip->dac_ops->output(chip, DEVICE_ENABLE);
    forza_print_runtime_info(substream->runtime,"START");
    if(PLAYBACK_TRIGGER_DELAY == 0){
      FIFO_CONTROL_REG_W(chip,1); // FIFO Clear
      chip->dma_ops->start(chip);
      FORZA_DEBUG_LOG("DMA Begining.....Delay Zero");
    }
    break;
  case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
    fpcm->running = 1;
    FORZA_DEBUG_LOG("PAUSE_RELEASE.....");
    break;
  case SNDRV_PCM_TRIGGER_RESUME:
    fpcm->running = 1;
    FORZA_DEBUG_LOG("RESUME....");
    break;
  case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
    fpcm->running = 0;
    FORZA_DEBUG_LOG("PAUSE_PUSH.....");
    break;
  case SNDRV_PCM_TRIGGER_SUSPEND:
    fpcm->running = 0;
    FORZA_DEBUG_LOG("SUSPEND.....");
    break;
  case SNDRV_PCM_TRIGGER_STOP:
    chip->audio_ops->mute(chip,DEVICE_ENABLE);
    chip->IsPlayReq = DEVICE_DISABLE;
    FORZA_INFO_LOG("TRG.dsd.stp...\n");
    udelay(800);
    //    FIFO_CONTROL_REG_W(chip,1);
    fpcm->running = 0;
    fpcm->irq_called_cnt = 0;
    fpcm->dma_period_cnt = 0;
    break;
  default:
    FORZA_DEBUG_LOG(">>DEFAULT\n");
    return -EINVAL;
  }
  return 0;
}


// open callback
static int snd_forza_capture_open(struct snd_pcm_substream *substream)
{
  struct snd_forza *chip = snd_pcm_substream_chip(substream);
  chip->substream = substream;
  return 0;
}

// close callback
static int snd_forza_capture_close(struct snd_pcm_substream *substream)
{
  return 0;
}


/* PCM Playback */
static struct snd_pcm_ops snd_forza_playback_ops = {
  .open      = snd_forza_playback_open,
  .close     = snd_forza_playback_close,
  .ioctl     = snd_pcm_lib_ioctl,
  .hw_params = snd_forza_playback_hw_params,
  .hw_free   = snd_forza_playback_hw_free,
  .prepare   = snd_forza_playback_prepare,
  .trigger   = snd_forza_pcm_trigger,
  .pointer   = snd_forza_playback_pointer,
};

/* DSD Playback */
static struct snd_pcm_ops snd_forza_dsd_playback_ops = {
  .open      = snd_forza_dsd_playback_open,
  .close     = snd_forza_dsd_playback_close,
  .ioctl     = snd_pcm_lib_ioctl,
  .hw_params = snd_forza_dsd_playback_hw_params,
  .hw_free   = snd_forza_dsd_playback_hw_free,
  .prepare   = snd_forza_dsd_playback_prepare,
  .trigger   = snd_forza_dsd_pcm_trigger,
  .pointer   = snd_forza_playback_pointer,
};

static struct snd_pcm_ops snd_forza_capture_ops = {
  .open =   snd_forza_capture_open,
  .close = snd_forza_capture_close,
};

// create a pcm device 
int __devinit snd_forza_new_pcm(struct snd_forza *chip, int device)
{
  struct snd_pcm *pcm;
  int err;
  err = snd_pcm_new(chip->card, "forzaFPGA", device, 1, 0, &pcm);
  if(err < 0){
    FORZA_ERROR_LOG("Can't create new pcm");
    return err;
  }
  pcm->private_data = chip;
  strncpy(pcm->name, "PCM(44.1K-192K)/32", sizeof(pcm->name));
  chip->pcm = pcm;
  snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_forza_playback_ops);
  err = snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV_SG,
					      snd_dma_pci_data(chip->pci),
					      128*5*1024, 128*5*1024);

  if(err < 0){
    FORZA_ERROR_LOG("Can't get PCM_LIB_PREALLOCATE_PAGE_FOR_ALL");
    return err;
  }
  return 0;
}

int __devinit snd_forza_new_pcm_dsd(struct snd_forza *chip, int device ){

  struct snd_pcm *pcm;
  int err;
  err = snd_pcm_new(chip->card, "forzaFPGA", device, 1 , 0 , &pcm);
  if(err < 0){
    FORZA_ERROR_LOG("Can't create new pcm");
    return err;
  }
  pcm->private_data = chip;
  strncpy(pcm->name, "DSD(2.8,5.6M)/32",sizeof(pcm->name));
  chip->pcm = pcm;
  snd_pcm_set_ops(pcm, SNDRV_PCM_STREAM_PLAYBACK, &snd_forza_dsd_playback_ops);
  err = snd_pcm_lib_preallocate_pages_for_all(pcm, SNDRV_DMA_TYPE_DEV_SG,
					      snd_dma_pci_data(chip->pci),
			      		      128*5*1024, 128*5*1024);
  if(err < 0){
    FORZA_ERROR_LOG("Can't get PCM_LIB_PREALLOCATE_PAGE_FOR_ALL");
    return err;
  }
  return 0;
}
